#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <sps2lib.h>
#include "sps2util_Godai.h"
#include "DJR_Mixer.h"
#include "DJR_Audiosystem.h"
#include "DJR_FrameMemMngr.h"
#include "DJR_Timers.h"
#include "UsefulMacros.h"
#include "fontAdobe.h"
#include "DJR_PadManager.h"

// prototypes
Bool MixerCreateNewFragment( void );
Bool MixerCheckEvents( void );

// statics
static short *pSampleBuffer[ 2 ];
static SampleState State[ 2 ];
static void *pMemFramePtr[ 2 ];
static short m_pFragment[ 512 ];

MixerEventCallbackFunc MixerEventCallbacks[ MIXER_EVENT_MAX ];

/*******************************************/
void MixerInit( void )
{
  int iCallbackLoop;
  pSampleBuffer[ 0 ] = pSampleBuffer[ 1 ] = NULL;
  pMemFramePtr[ 0 ] = pMemFramePtr[ 1 ] = NULL;
  // all we need to do is initialize the state
  State[ 0 ].fNumSamples = State[ 1 ].fNumSamples = 0.f;
  State[ 0 ].fSpeed = State[ 1 ].fSpeed = 1.f;
  State[ 0 ].fCurPos = State[ 1 ].fCurPos = 0.f;
  State[ 0 ].bLoop = State[ 1 ].bLoop = FALSE;
  State[ 0 ].bLoaded = State[ 1 ].bLoaded = FALSE;
  State[ 0 ].bPlaying = State[ 1 ].bPlaying = FALSE;
  State[ 0 ].fInitialBPM = State[ 1 ].fInitialBPM = 0.f;

  // reset all func pointerz, yo
  for( iCallbackLoop=0; iCallbackLoop<MIXER_EVENT_MAX; iCallbackLoop++ )
  {
    MixerEventCallbacks[ iCallbackLoop ] = NULL;
  }
}

/*********************************************/
void MixerKill( void )
{
    FrameMemMngrDeallocateAll( STACKTYPE_T1, TRUE );
    FrameMemMngrDeallocateAll( STACKTYPE_T1, FALSE );
    FrameMemMngrDeallocateAll( STACKTYPE_T2, TRUE );
    FrameMemMngrDeallocateAll( STACKTYPE_T2, FALSE );
}

/**********************************************/
Bool MixerLoadSample( char *pFileName, char iTable, 
                      float fSpeed, Bool bLoop )
{
    Bool bSuccess = FALSE;
    struct stat FileInfo;
    if( stat( pFileName, &FileInfo) < 0 )
       return FALSE;

    // turntable 2 and 1 use different memory stacks
    StackType eStackType = iTable == 0?
                             STACKTYPE_T1 :
                             STACKTYPE_T2;

    // get a frame so we can free later
    if( State[ iTable ].bLoaded )
    {
        // if there is already a record loaded on this table
        // then we must have a frame to free 
        //SYS_ASSERT( pMemFramePtr[ iTable ] );
        FrameMemMngrDeallocateFrame( eStackType,
                                     TRUE, 
                                     pMemFramePtr[ iTable ] );
    }

    // obtain a new frame so we can deallocate later
    pMemFramePtr[ iTable ] = 
        FrameMemMngrObtainFrame( eStackType, TRUE );

    // actually allocate the memory
    pSampleBuffer[ iTable ] = ( short * ) 
        FrameMemMngrAllocate( eStackType, TRUE, FileInfo.st_size );

    if( pSampleBuffer[ iTable ] )
    {
        // load the data into the buffer we just allocated
        int fh;
        if( ( fh = open( pFileName, O_RDONLY ) ) >= 0 )
	{ 
	    int iBytesRead;
	    iBytesRead = read( fh, pSampleBuffer[ iTable ], 
                               FileInfo.st_size );
            close( fh );
            // setup the sample state
            State[ iTable ].fNumSamples = (float)( FileInfo.st_size / 4 );
            State[ iTable ].fSpeed = fSpeed;
            State[ iTable ].fCurPos = 
                fSpeed > 0.f? 0.f : State[ iTable ].fNumSamples - 1.f;
            State[ iTable ].bLoop = bLoop;
            State[ iTable ].bLoaded = TRUE;
            State[ iTable ].bPlaying = FALSE;
            bSuccess = TRUE;
            if( bLoop )
	    {
	        // sample length in seconds
	        float fTmpBPM = State[ iTable ].fNumSamples / 48000;
                // BPM assuming sample is 2 measures (8 beats) long in 4/4
                fTmpBPM = 8 * 60 / fTmpBPM;
	        State[ iTable ].fInitialBPM = fTmpBPM;
	    }
            else
	    {
	        State[ iTable ].fInitialBPM = -1.f;
            }
	}
    }

    // todo: clean up memory if we failed to load the file
    if( !bSuccess )
    {
        State[ iTable ].bLoaded = FALSE;
        // if we allocated memory but failed to open the file, deallocate
        if( pSampleBuffer[ iTable ] )
	{
	    FrameMemMngrDeallocateFrame( eStackType,
                                         TRUE,
                                         pMemFramePtr[ iTable ] );
            printf( "\nfailed to open file %s", pFileName );
	}
        else
	    printf( "\nsomething went wrong allocating memory for %s", pFileName );
        fflush( NULL );
        return FALSE;
    }
    printf( "\n\nsuccessfully loaded %s (%d bytes) on table %d\n", 
                  pFileName, FileInfo.st_size, iTable );
    fflush( NULL );
    return TRUE;
}

/***********************************************/
void MixerUnloadSample( char iTurntable )
{

}

/***********************************************/
Bool MixerTogglePause( char iTurntable )
{
  return State[ iTurntable ].bPlaying = !State[ iTurntable ].bPlaying;
}

/************************************************/
Bool MixerProcess( Bool bReset, Bool bTimerExpired )
{
  Bool bExit = FALSE;
  // time to make a new fragment???
  // stop the system from blocking

  if( bTimerExpired )
  {
      PadUpdatePadData( );
      //if( PadIsButtonPressed( 0, PS2PAD_BUTTON_START ) )
      //    bExit = TRUE;
  }

  if( OSSDeviceMaxBytes() >= 1024 )
  {
    MixerCheckEvents( );
    MixerCreateNewFragment( );

    // thrown in to test how speed interp sounds
    State[ 0 ].fSpeed += 0.0002f;
    State[ 1 ].fSpeed += 0.0002f;
  }
  return bExit;
}

/**************************************************/
Bool MixerCreateNewFragment( void )
{
  // no records loaded...  nothing to do
  if( !State[ 0 ].bLoaded && !State[ 1 ].bLoaded )
    return FALSE;  

  // if we only have one record loaded, set some pointers to avoid crashing
  static short iNotLoadedBuf[ 2 ] = { 0, 0 };
  if( !pSampleBuffer[ 0 ] )
      pSampleBuffer[ 0 ] = &iNotLoadedBuf;
  else if( !pSampleBuffer[ 1 ] )
      pSampleBuffer[ 1 ] = &iNotLoadedBuf;

  // fragment size is 1024 and stereo, 
  // so we have 512 bytes of audio to generate
  // each sample is is a signed short so we generate 256 samples
  unsigned int uSamplesPerFrag = GetFragSize( ) / 2 / 2;

  // TODO: get volumes from the mixer volume and crossfader position
  float fVolume[ 2 ] = { 0.5f, 0.5f };
  int iFragBufPtr = 0;

  while( uSamplesPerFrag > 0 )
  {   
    // mix samples.  fCurPos is mul * 2 because its stereo data
    short iSample = ( short )( pSampleBuffer[ 0 ][ 2 * ROUND( State[ 0 ].fCurPos ) ] * fVolume[ 0 ] );
    iSample += ( short )( pSampleBuffer[ 1 ][ 2 * ROUND( State[ 1 ].fCurPos ) ] * fVolume[ 1 ] );

    // send the stereo sample to the DSP
    m_pFragment[ iFragBufPtr++ ] = iSample;
    m_pFragment[ iFragBufPtr++ ] = iSample;

    // dont write two shorts at a time...  write the whole buffer later
    //WriteSampleToFragment( &m_pFragment[ iFragBufPtr - 2 ] );

    // update sample pointers if the sample isnt paused
    // TODO: play around with rewriting this in assembly 
    int uTable;
    for( uTable=0; uTable<2; uTable++ )
    {
      // if this turntable has a song loaded and its not paused...
      if( State[ uTable ].bLoaded && State[ uTable ].bPlaying )
      {
	State[ uTable ].fCurPos += State[ uTable ].fSpeed;
                    
        // if this sample is supposed to loop
        if( State[ uTable ].bLoop )
	{
	  // going backwards and now at the beginning, so jump to the end
	  if( State[ uTable ].fCurPos <= 0.f )
	    State[ uTable ].fCurPos = State[ uTable ].fNumSamples - 1.f;
                        
          // going forward and now at the end, so go to the beginning
	  else if( State[ uTable ].fCurPos>=State[ uTable ].fNumSamples )
	    State[ uTable ].fCurPos = 0.f;
	}
	else
	{
	  // if we're not looping, clamp to the sample size 
	  if( State[ uTable ].fCurPos < 0 )
	  {
              State[ uTable ].fCurPos = 0;
              State[ uTable ].bPlaying = FALSE;
	  }
          else if( State[ uTable ].fCurPos > State[ uTable ].fNumSamples - 1.f )
	  {
              State[ uTable ].fCurPos = State[ uTable ].fNumSamples - 1;
              State[ uTable ].bPlaying = FALSE;
	  }
	}
      }
    }
    --uSamplesPerFrag;  
  }

  WriteBuffToDSP( m_pFragment );

  return FALSE;
}

/**************************************************************/
void MixerRenderDebugText( void )
{
  int iTable;
  for( iTable=0; iTable<2; iTable++ )
  {
    int iXpos = 100 + 300 * iTable; 
    sps2UPrintf( 0, "\376(%i,10)turntable %d", iXpos, iTable );
    sps2UPrintf( 0, "\376(%i,30)track name: %s", iXpos, "none" );
    sps2UPrintf( 0, "\376(%i,50)speed: %f", iXpos, State[iTable].fSpeed );
    sps2UPrintf( 0, "\376(%i,70)loop: %s", 
        iXpos,
	State[ iTable ].bLoop? "yes" : "no" );
    sps2UPrintf( 0, "\376(%i,90)playing: %s",
        iXpos,
        State[ iTable ].bPlaying? "yes" : "no" );
    sps2UPrintf( 0, "\376(%i,110)position: %ld / %ld",
	iXpos,
	( long int )State[ iTable ].fCurPos, 
	( long int )State[ iTable ].fNumSamples );
    if( State[ iTable ].bLoop )
    {
        sps2UPrintf( 0, "\376(%i,130)approx loop BPM: %f", 
            iXpos, State[ iTable ].fInitialBPM * State[ iTable ].fSpeed );
    }
    else
    {
        sps2UPrintf( 0, "\376(%i,130)approx loop BPM: not a loop", iXpos );
    }
  }
  sps2UPrintfRender( );
}

/*******************************************************************/
float MixerGetRecordSpeed( char iTurntable )
{
  // this is the speed multiplier, NOT the BPM

  // if the record isnt loaded or isnt playing, ignore the speed
  // TODO: return 0 for non-looping records at the end
  if( !State[ iTurntable ].bLoaded ||
      !State[ iTurntable ].bPlaying )
  {
      return( 0.f );
  }

  else
      return( State[ iTurntable ].fSpeed );
}

/************************************************************************/
Bool MixerCheckEvents( void )
{ 
    Bool bEventOccured = FALSE;
    int iMixerEvent;
    int iTable;

    // look for events on both turntables
    for( iTable=0; iTable<2; iTable++ )
    {
        // no events for unloaded records, or ones not currently playing
        if( !State[ iTable ].bLoaded || !State[ iTable ].bPlaying )
	    continue;

        // look through all events to see what happened
        for( iMixerEvent=0; iMixerEvent<MIXER_EVENT_MAX; iMixerEvent++ )
        {
            // no need to check this event if there is no callback for it
            if( !MixerEventCallbacks[ iMixerEvent ] )
                continue;

            // do we want to call this event's callback function?
            Bool bDoCallback = FALSE;

            // get the current and next sample ptr just to save code space 
            float fCur = State[ iTable ].fCurPos;
            float fNxt = fCur + 1024 / 4 * State[ iTable ].fSpeed;
            float fTgt;

            // look for each event and see if it occured
            switch( iMixerEvent )
            {
                // another 1/4th of the loop has played
                case MIXER_EVENT_LOOP_QUARTER:
                {
		    int iTgtLoop;

                    // no loop events for non-looping samples
                    if( !State[ iTable ].bLoop )
		        break;

                    for( iTgtLoop=0; iTgtLoop<=4; iTgtLoop++ )
		    {
		        // target is 0/4, 1/4, 2/4, 3/4, 4/4 of NumSanples 
		        fTgt = iTgtLoop / 4.f;
                        fTgt *= State[ iTable ].fNumSamples;
                        if( BETWEEN( fCur, fTgt, fNxt ) || 
                            BETWEEN( fNxt, fTgt, fCur ) )
	                {
                            bEventOccured = TRUE;
                            bDoCallback = TRUE;
                            break;
                        }
		    }     
                }
                break;

                default:
                    // invalid event
                    break;
            }  

            // if we wanna do the callback, do it now!  get in the chopper!
            if( bDoCallback )
                MixerEventCallbacks[ iMixerEvent ]( iTable, NULL );
        }
    }   

    return bEventOccured;
}

/*********************************************************************/
void MixerEventSetCallback( MixerEvent eEvent,
                            MixerEventCallbackFunc pCallbackFunc )
{
  // set the callback for ALL events
  if( eEvent == MIXER_EVENT_ALL )
  {
    int iEventCnt;
    for( iEventCnt=0; iEventCnt<MIXER_EVENT_MAX; iEventCnt++ )
      MixerEventCallbacks[ iEventCnt ] = pCallbackFunc;
  }
  else
  {
    MixerEventCallbacks[ eEvent ] = pCallbackFunc;
  }

  printf( "\n\nadding callback for event %d (func = 0x%X)",
      eEvent, pCallbackFunc );
}
