#ifndef _DJR_MIXER_H_
#define _DJR_MIXER_H_

#include "UsefulMacros.h"

typedef Bool (*MixerEventCallbackFunc)(int, void *);

// types of mixer events that can have callbacks
// for simplicity, each event can have ONE callback
typedef enum 
{
  MIXER_EVENT_LOOP_NEW,
  MIXER_EVENT_LOOP_HALF,
  MIXER_EVENT_LOOP_QUARTER,
  MIXER_EVENT_LOOP_EIGHTH,
  MIXER_EVENT_LOOP_PAUSE,
  MIXER_EVENT_NEW_RECORD,
  MIXER_EVENT_NEW_TRACK,
  MIXER_EVENT_ALL,
  MIXER_EVENT_MAX = MIXER_EVENT_ALL
}MixerEvent;

// keep track of each turntable
typedef struct
{
  float fNumSamples; // file size / 2 channels / 2 bytes per sample
  float fSpeed;  // 1=normal, 2=double, 0.5=half, -1=normal backwards
  float fCurPos; // current sample.  x2 (stereo) for the real position
  float fInitialBPM; // assumes loop is 2 measures (8 beats) in 4/4 time
  Bool bLoop; // loop or one-shot
  Bool bLoaded; // do we have a valid sample loaded
  Bool bPlaying; // playing or paused
}SampleState;

void MixerInit( void );
void MixerKill( void );
Bool MixerLoadSample( char *pFileName, char iTable, 
                      float fSpeed, Bool bLoop );
void MixerUnloadSample( char iTurntable );
Bool MixerTogglePause( char iTurntable );
Bool MixerProcess( Bool bReset, Bool bTimerExpired );
void MixerRenderDebugText( void );
float MixerGetRecordSpeed( char iTurntable );
void MixerEventSetCallback( MixerEvent eEvent, 
                            MixerEventCallbackFunc pCallbackFunc );
 
#endif
