#include <math.h>
#include <stdio.h>
#include <sps2tags.h>
#include "DJR_SimpleTexMngr.h"
#include "DJR_Model.h"
#include "gsregs.h"
#include "UsefulMacros.h"
  
// *******************************
InitModel( StaticModel_t *pModel, int iSPS2Descriptor )
{
  pModel->m_bLoaded = FALSE;
  pModel->m_iNumPages = 0;
  pModel->m_iTextureSlot = -1;
  pModel->m_iSPS2Descriptor = iSPS2Descriptor;
  pModel->m_pDMAChainPages = NULL;	
}

// *******************************
bool LoadDMAFile( StaticModel_t *pModel, char *pFileName, int iTexSlot )
{
  bool bSuccess = FALSE;
  struct stat FileInfo;
  int fh = -1;
	
  // get the file size
  if( stat( pFileName, &FileInfo) < 0 )
    return FALSE;
	
  // open the file 
  if( ( fh = open( pFileName, O_RDONLY ) ) == -1 )
    return FALSE;
	
  printf( "\n\nloading static model %s (%f QW)", 
    pFileName, FileInfo.st_size / 16.f );
	
  // make sure we dont already have a mode loaded
  DumpModel( pModel );
	
  // figure out how many pages (4096) we need to store this model's chain
  // not including any needed dma tags except for the first one
	
  // dma tag ( 8 bytes ) + giftag ( 16 bytes )+ texflush ( 16 bytes ) + 
  // tex0_1 ( 16 bytes ) +
  // prim (16 bytes) + FileInfo.st_size
  pModel->m_iNumPages = 8 + 16 + 16 + 16 + 16 + FileInfo.st_size;
  pModel->m_iNumPages = (int)ceilf( pModel->m_iNumPages / 4096.f );
  // now figure out how many more DMA tags we need for n pages
  pModel->m_iNumPages = ( pModel->m_iNumPages * 8 ) + 8 + 16 + 
                        16 + 16 + 16 + FileInfo.st_size;
  pModel->m_iNumPages = (int)ceilf( pModel->m_iNumPages / 4096.f );
  printf( "\nusing %d 4096 byte pages", pModel->m_iNumPages );

/*

  // allocate some page pointers and then the pages themselves
  m_pDMAChainPages = ( sps2Memory_t ** )malloc( m_iNumPages );

  // allocate all the needed pages
  int iPageCnt;
  for( iPageCnt = 0; iPageCnt<m_iNumPages; iPageCnt++ )
  {
    m_pDMAChainPages[ iPageCnt ] = sps2Allocate( 4096, 
                                     SPS2_MAP_BLOCK_4K | SPS2_MAP_UNCACHED,
				     m_iSPS2Descriptor );    	
  }

  // load the first page and insert the initial dmatag, 
  // giftag and registers
  sps2DMATag_t *pDMATag = ( sps2DMATag_t * )
                              ( m_pDMAChainPages[ 0 ]->pvStart );
  pDMATag->s.PCE = 0; 
  pDMATag->s.SPR = 0;   // main mem, not scratchpad
  pDMATag->s.IRQ = 0;    // no interrupt

  if( m_iNumPages > 1 )
  {
    pDMATag->s.QWC = 255;
    pDMATag->s.ID = DMA_ID_SOURCE_NEXT;
    pDMATag->s.ADDR = ( sps2uint32 )( m_pDMAChainPages[ 1 ]->pvStart );
  }
  else
  {
    pDMATag->s.QWC = ( 8 + 16 + 16 + 16 + 16 + FileInfo.st_size ) / 16; 
    pDMATag->s.ID = DMA_ID_SOURCE_END;
    pDMATag->s.ADDR = 0;
  }
  pDMATag++;

  sps2GIFTag_t *gt = (sps2GIFTag_t *)pDMATag;
  sps2uint64 *regs = (sps2uint64 *) ((char*)gt+sizeof(sps2GIFTag_t));

  gt->i128 = 0;  
  gt->s.NLOOP = 1; 
  gt->s.EOP = 0;
  gt->s.PRE = 0;   
  gt->s.PRIM = 0;
  gt->s.FLG = GIF_FLG_PACKED;
  gt->s.NREG = 0;

  // add an A_D register
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // TEXFLUSH
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // TEX0_1 
  SPS2_SET_GIF_REG(*gt,gt->s.NREG++,GIF_REG_A_D);  // PRIM

  // access the texflush register with anything
  *(regs++) = (sps2uint64)0xAA55;
  *(regs++) = TEXFLUSH_R;
  sps2uint32 tex_mem_addr = g_pTexMngr->GetSlotGSAddress( iTexSlot );
  *(regs++) = PACK_TEX0(tex_mem_addr/64,TEX_W/64,0,8,8,1,1,0,0,0,0,0);
  *(regs++) = TEX0_1_R;
  *(regs++) = PACK_PRIM(3,1,1,0,0,1,0,0,0);
  *(regs++) = PRIM_R;

  // try to read enough to fill the page
  // on the first page, 80 bytes are used for dma tags, 
  // the gif tag, and regs
  int iBytesRead = read( fh, regs, 4096 - 80 );
	
  // load the rest of the pages
  for( iPageCnt = 1; iPageCnt<m_iNumPages; iPageCnt++ )
  {
    pDMATag = ( sps2DMATag_t * )( m_pDMAChainPages[ iPageCnt ]->pvStart );
    if( iPageCnt < m_iNumPages - 1 )
    {
      pDMATag->s.ID = DMA_ID_SOURCE_NEXT;
      pDMATag->s.ADDR = ( sps2uint32 ) 
                            (m_pDMAChainPages[ iPageCnt + 1 ]->pvStart );
    }
    else
    {
      pDMATag->s.ID = DMA_ID_SOURCE_END;
      pDMATag->s.ADDR = 0;
    }
    iBytesRead = read( fh, ( void * )( pDMATag + 1 ), 4096 - 16 );
    pDMATag->s.QWC = iBytesRead / 16;
  }
	
  close( fh );
*/
  return TRUE;
}

// *******************************
void DumpModel( StaticModel_t *pModel )
{
  pModel->m_iTexSlot = -1;

  if( pModel->m_iNumPages > 0 && pModel->m_pDMAChainPages )
  {
    do
    {
      --( pModel->m_iNumPages );
      sps2Free( pModel->m_pDMAChainPages[ pModel->m_iNumPages ] );
    }while( pModel->m_iNumPages > 0 );
		
    free( pModel->m_pDMAChainPages );
    pModel->m_pDMAChainPages = NULL;	
  }
  pModel->m_bLoaded = FALSE;
}

// *******************************
void RenderModel( StaticModel_t *pModel )
{
	
}

