#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sps2lib.h>
#include "sps2util_Godai.h"
#include "UsefulMacros.h"
#include "DJR_Timers.h"
#include "DJR_Audiosystem.h"
#include "DJR_Mixer.h"
#include "DJR_FrameMemMngr.h"
#include "DJR_VsyncIntHandle.h"
#include "DJR_SimpleTexMngr.h"
#include "DJR_PadManager.h"

int main( int argc, char **argv )
{
  printf( "\n\nfree memory before game init:\n" );
  fflush( NULL );
  system( "free" );

  // init the dualshock2 stuff
  PadInit( );

  AudiosystemInit( );

  // init sps2 for direct hardware access
  int iSPS2Device = sps2Init( );
  if( iSPS2Device < 0 )
  {
    exit( -1 );
  }

  // supervisor mode messes up my vsync detection.  DON'T USE IT!!!
  //sps2SetOperationMode( 1, iSPS2Device );

  // init the screen
  sps2UScreenInit( 0 );

  SimpleTexMngrInit( iSPS2Device );
  FrameMemMngrInit( );
  MixerInit( );

  FrameMemMngrCreate( 1250000, 64 );

  struct timeval sRandomSeed;
  gettimeofday( &sRandomSeed,NULL );
  srandom( sRandomSeed.tv_usec );

  // load 2 records and start playing them
  MixerLoadSample("sample-audio/DeadBuffalos.raw", 0, 1.f, TRUE);
  MixerTogglePause( 0 );
  MixerLoadSample("sample-audio/shoryuken.raw", 1, 1.f, FALSE);
  MixerTogglePause( 1 ); 

  // must be done AFTER initializing tbe texture manager
  InstallVsyncInterruptHandler( iSPS2Device );

  SetTimer( 1.f / 30.f );

  unsigned long iTestLoop;
  for( iTestLoop=0; iTestLoop<4500000; iTestLoop++ )
  {
      Bool bTimerExpired = DidTimerExpire( );
      if( MixerProcess( FALSE, bTimerExpired ) )
	  break;

      if( bTimerExpired )
          SetTimer( 1.f / 30.f );
  }

  RemoveVsyncInterruptHandler( );

  FrameMemMngrDestroy( );

  // must be done IN ORDER!!!
  SimpleTexMngrKill( );
  MixerKill( );
  FrameMemMngrKill( );

  sps2UScreenShutdown( );
  sps2SetOperationMode( 0, iSPS2Device );
  sps2Release( iSPS2Device );
  PadShutdown( );

  //sleep( 5 );
  printf( "\n\nfree memory after game shutdown:\n" );
  fflush( NULL );
  system( "free" );

  return 0;
}
